import dataclasses
import inspect
import sys
import typing
from abc import ABC, abstractmethod, ABCMeta
from dataclasses import Field, MISSING, dataclass, field
from typing import Optional, Callable, List, Union

from .naming import ReferenceByName
from .position import Position, Source
from .reflection import Multiplicity, PropertyDescription
from ..reflection import get_type_annotations, get_type_arguments, is_sequence_type
from ..reflection.reflection import get_type_origin

PYLASU_FEATURE = "pylasu_feature"


class internal_property(property):
    pass


def internal_properties(*props: str):
    def decorate(cls: type):
        cls.__internal_properties__ = (
            getattr(cls, "__internal_properties__", []) + [*Node.__internal_properties__, *props])
        return cls
    return decorate


class InternalField(Field):
    pass


def internal_field(
        *, default=MISSING, default_factory=MISSING, init=True, repr=True, hash=None, compare=True, metadata=None,
        kw_only=False):
    """Return an object to identify internal dataclass fields. The arguments are the same as dataclasses.field."""

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')
    try:
        # Python 3.10+
        return InternalField(default, default_factory, init, repr, hash, compare, metadata, kw_only)
    except TypeError:
        return InternalField(default, default_factory, init, repr, hash, compare, metadata)


def node_property(default=MISSING):
    description = PropertyDescription(
        "", None,
        multiplicity=Multiplicity.OPTIONAL if default is None else Multiplicity.SINGULAR)
    return field(default=default, metadata={PYLASU_FEATURE: description})


def node_containment(multiplicity: Multiplicity = Multiplicity.SINGULAR):
    description = PropertyDescription("", None, is_containment=True, multiplicity=multiplicity)

    if multiplicity == Multiplicity.SINGULAR:
        return field(metadata={PYLASU_FEATURE: description})
    elif multiplicity == Multiplicity.OPTIONAL:
        return field(default=None, metadata={PYLASU_FEATURE: description})
    elif multiplicity == Multiplicity.MANY:
        return field(default_factory=list, metadata={PYLASU_FEATURE: description})


class Origin(ABC):
    @internal_property
    @abstractmethod
    def position(self) -> Optional[Position]:
        pass

    @internal_property
    def source_text(self) -> Optional[str]:
        return None

    @internal_property
    def source(self) -> Optional[Source]:
        return self.position.source if self.position is not None else None


@dataclass
class CompositeOrigin(Origin):
    elements: List[Origin] = field(default_factory=list)
    position: Optional[Position] = None
    source_text: Optional[str] = None


class Destination(ABC):
    pass


@dataclass
class CompositeDestination(Destination):
    elements: List[Destination] = field(default_factory=list)


@dataclass
class TextFileDestination(Destination):
    position: Optional[Position] = None


def is_internal_property_or_method(value):
    return isinstance(value, internal_property) or isinstance(value, InternalField) or isinstance(value, Callable)


def provides_nodes(decl_type):
    if get_type_origin(decl_type) is Union:
        provides = None
        for tp in get_type_arguments(decl_type):
            if tp is type(None):
                continue
            arg_provides = provides_nodes(tp)
            if provides is None:
                provides = arg_provides
            elif provides != arg_provides:
                raise Exception(f"Type {decl_type} mixes nodes and non-nodes")
        return provides
    else:
        return isinstance(decl_type, type) and issubclass(decl_type, Node)


def get_only_type_arg(decl_type):
    """If decl_type has a single type argument, return it, otherwise return None"""
    type_args = get_type_arguments(decl_type)
    if len(type_args) == 1:
        return type_args[0]
    else:
        return None


def process_annotated_property(cl: type, name: str, decl_type):
    try:
        fields = dataclasses.fields(cl)
    except TypeError:
        fields = tuple()
    for fld in fields:
        if fld.name == name and PYLASU_FEATURE in fld.metadata:
            feature = fld.metadata[PYLASU_FEATURE]
            feature.name = name
            if isinstance(decl_type, type):
                feature.type = decl_type
            elif type(fld.type) is str:
                feature.type = try_to_resolve_string_type(fld.type, name, cl)
            return feature
    return compute_feature_from_annotation(cl, name, decl_type)


def compute_feature_from_annotation(cl, name, decl_type):
    feature = PropertyDescription(name, None, False, False, Multiplicity.SINGULAR)
    decl_type = try_to_resolve_type(decl_type, feature)
    if not isinstance(decl_type, type):
        fwref = None
        if hasattr(typing, "ForwardRef"):
            fwref = typing.ForwardRef
        if fwref and isinstance(decl_type, fwref):
            raise Exception(f"Feature {name}'s type is unresolved forward reference {decl_type}, "
                            f"please use node_containment or node_property")
        elif type(decl_type) is str:
            decl_type = try_to_resolve_string_type(decl_type, name, cl)
        if not isinstance(decl_type, type):
            raise Exception(f"Unsupported feature {name} of type {decl_type}")
    feature.type = decl_type
    feature.is_containment = provides_nodes(decl_type) and not feature.is_reference
    return feature


def try_to_resolve_string_type(decl_type, name, cl):
    try:
        ns = getattr(sys.modules.get(cl.__module__, None), '__dict__', globals())
        decl_type = ns[decl_type]
    except KeyError:
        raise Exception(f"Unsupported feature {name} of unknown type {decl_type}")
    return decl_type


def try_to_resolve_type(decl_type, feature):
    if get_type_origin(decl_type) is ReferenceByName:
        decl_type = get_only_type_arg(decl_type) or decl_type
        feature.is_reference = True
    if is_sequence_type(decl_type):
        decl_type = get_only_type_arg(decl_type) or decl_type
        feature.multiplicity = Multiplicity.MANY
    if get_type_origin(decl_type) is Union:
        type_args = get_type_arguments(decl_type)
        if len(type_args) == 1:
            decl_type = type_args[0]
        elif len(type_args) == 2:
            if type_args[0] is type(None):
                decl_type = type_args[1]
            elif type_args[1] is type(None):
                decl_type = type_args[0]
            else:
                raise Exception(f"Unsupported feature {feature.name} of union type {decl_type}")
            if feature.multiplicity == Multiplicity.SINGULAR:
                feature.multiplicity = Multiplicity.OPTIONAL
        else:
            raise Exception(f"Unsupported feature {feature.name} of union type {decl_type}")
    return decl_type


class Concept(ABCMeta):

    def __init__(cls, what, bases=None, dict=None):
        super().__init__(what, bases, dict)
        cls.__internal_properties__ = []
        for base in bases:
            if hasattr(base, "__internal_properties__"):
                cls.__internal_properties__.extend(base.__internal_properties__)
        if not cls.__internal_properties__:
            cls.__internal_properties__ = ["origin", "destination", "parent", "position", "position_override"]
        cls.__internal_properties__.extend([n for n, v in inspect.getmembers(cls, is_internal_property_or_method)])

    @property
    def node_properties(cls):
        names = set()
        for cl in cls.__mro__:
            yield from cls._direct_node_properties(cl, names)

    def _direct_node_properties(cls, cl, known_property_names):
        if not isinstance(cls, Concept):
            return
        anns = get_type_annotations(cl)
        if not anns:
            return
        for name in anns:
            if name not in known_property_names and cls.is_node_property(name):
                feature = process_annotated_property(cl, name, anns[name])
                known_property_names.add(name)
                yield feature
        for name in dir(cl):
            if name not in known_property_names and cls.is_node_property(name):
                feature = PropertyDescription(name, None, False, False)
                known_property_names.add(name)
                yield feature

    def is_node_property(cls, name):
        return not name.startswith('_') and name not in cls.__internal_properties__


class Node(Origin, Destination, metaclass=Concept):
    origin: Optional[Origin] = None
    destination: Optional[Destination] = None
    parent: Optional["Node"] = None
    position_override: Optional[Position] = None

    def __init__(self, origin: Optional[Origin] = None, parent: Optional["Node"] = None,
                 position_override: Optional[Position] = None):
        self.origin = origin
        self.parent = parent
        self.position_override = position_override

    def with_origin(self, origin: Optional[Origin]):
        self.origin = origin
        return self

    def with_parent(self, parent: Optional["Node"]):
        self.parent = parent
        return self

    def with_position(self, position: Optional[Position]):
        self.position = position
        return self

    @internal_property
    def position(self) -> Optional[Position]:
        return self.position_override if self.position_override is not None\
            else self.origin.position if self.origin is not None else None

    @position.setter
    def position(self, position: Optional[Position]):
        self.position_override = position

    @internal_property
    def source_text(self) -> Optional[str]:
        return self.origin.source_text if self.origin is not None else None

    @internal_property
    def source(self) -> Optional[Source]:
        return self.origin.source if self.origin is not None else None

    @internal_property
    def properties(self):
        return (PropertyDescription(p.name, p.type,
                                    is_containment=p.is_containment, is_reference=p.is_reference,
                                    multiplicity=p.multiplicity, value=getattr(self, p.name))
                for p in self.__class__.node_properties)

    @internal_property
    def _fields(self):
        yield from (name for name, _ in self.properties)

    @internal_property
    def node_type(self):
        return type(self)


def concept_of(node):
    properties = dir(node)
    if "__concept__" in properties:
        node_type = node.__concept__
    elif "node_type" in properties:
        node_type = node.node_type
    else:
        node_type = type(node)
    if isinstance(node_type, Concept):
        return node_type
    else:
        raise Exception(f"Not a concept: {node_type} of {node}")



import time
from abc import abstractmethod
from typing import Optional, List, Union

from antlr4 import CommonTokenStream, InputStream, Lexer, Parser, ParserATNSimulator, ParserRuleContext, \
    PredictionContextCache, Recognizer, Token, TokenStream
from antlr4.error.ErrorListener import ErrorListener
from pylasu.model import walk, Source, Position, Node, Point
from pylasu.model.processing import assign_parents
from pylasu.parsing.parse_tree import token_end_point
from pylasu.parsing.results import ParsingResultWithFirstStage, FirstStageParsingResult
from pylasu.validation import Issue, IssueType


class PylasuANTLRParser:
    """ A complete description of a multi-stage ANTLR-based parser, from source code to AST.

    You should extend this class to implement the parts that are specific to your language.

    Note: instances of this class are thread-safe and they're meant to be reused. Do not create a new PylasuANTLRParser
    instance every time you need to parse some source code, or performance may suffer."""
    def __init__(self):
        self.prediction_context_cache = PredictionContextCache()

    def parse(self, input_stream: Union[InputStream, str], consider_range: bool = True,
              measure_lexing_time: bool = False, source: Optional[Source] = None):
        """Parses source code, returning a result that includes an AST and a collection of parse issues
        (errors, warnings).
        The parsing is done in accordance to the StarLasu methodology i.e. a first-stage parser builds a parse tree
        which is then mapped onto a higher-level tree called the AST.
        @param inputStream the source code.
        @param charset the character set in which the input is encoded.
        @param considerPosition if true (the default), parsed AST nodes record their position in the input text.
        @param measureLexingTime if true, the result will include a measurement of the time spent in lexing i.e.
        breaking the input stream into tokens."""
        start = time.time_ns()
        if type(input_stream) is str:
            input_stream = InputStream(input_stream)
        first_stage = self.parse_first_stage(input_stream, measure_lexing_time)
        issues = first_stage.issues
        ast = self.parse_tree_to_ast(first_stage.root, consider_range, issues, source)
        self.assign_parents(ast)
        ast = self.post_process_ast(ast, issues) if ast else ast
        if ast and not consider_range:
            # Remove parseTreeNodes because they cause the range to be computed
            for node in walk(ast):
                node.origin = None
        now = time.time_ns()
        return ParsingResultWithFirstStage(
            issues,
            ast,
            input_stream.getText(0, input_stream.index + 1),
            (now - start) // 1_000_000,
            first_stage,
            source,
        )

    def parse_first_stage(
            self, input_stream: InputStream, measure_lexing_time: bool = False, source: Source = None
    ) -> FirstStageParsingResult:
        """Executes only the first stage of the parser, i.e., the production of a parse tree. Usually, you'll want to
        use the [parse] method, that returns an AST which is simpler to use and query."""
        issues = []
        lexing_time: Optional[int] = None
        total_time: int = time.time_ns()
        parser = self.create_parser(input_stream, issues)
        if measure_lexing_time:
            token_stream = parser.getInputStream()
            if isinstance(token_stream, CommonTokenStream):
                lexing_time = time.time_ns()
                token_stream.fill()
                token_stream.seek(0)
                lexing_time = (time.time_ns() - lexing_time) // 1_000_000
        root = self.invoke_root_rule(parser)
        if root:
            self.verify_parse_tree(parser, issues, root)
        total_time = (time.time_ns() - total_time) // 1_000_000
        return FirstStageParsingResult(issues, root, None, total_time, lexing_time, source)

    def create_parser(self, input_stream: InputStream, issues: List[Issue]) -> Parser:
        """Creates the first-stage parser."""
        lexer = self.create_antlr_lexer(input_stream)
        self.attach_listeners(lexer, issues)
        token_stream = self.create_token_stream(lexer)
        parser = self.create_antlr_parser(token_stream)
        # Assign interpreter to avoid caching DFA states indefinitely across executions
        parser._interp = \
            ParserATNSimulator(parser, parser.atn, parser._interp.decisionToDFA, self.prediction_context_cache)
        self.attach_listeners(parser, issues)
        return parser

    def invoke_root_rule(self, parser: Parser):
        """Invokes the parser's root rule, i.e., the method which is responsible for parsing the entire input.
        Usually this is the topmost rule, the one with index 0 (as also assumed by other libraries such as antlr4-c3),
        so this method invokes that rule. If your grammar/parser is structured differently, or if you're using this to
        parse only a portion of the input or a subset of the language, you have to override this method to invoke the
        correct entry point."""
        return getattr(parser, parser.ruleNames[0])()

    def verify_parse_tree(self, parser: Parser, issues: List[Issue], root: ParserRuleContext):
        """Checks the parse tree for correctness.
        If you're concerned about performance, you may want to override this to do nothing."""
        last_token: Token = parser.getTokenStream().get(parser.getTokenStream().index)
        if last_token.type != Token.EOF:
            issues.append(
                Issue(
                    IssueType.SYNTACTIC,
                    "The whole input was not consumed",
                    position=Position(token_end_point(last_token), token_end_point(last_token))
                )
            )
        # TODO Kolasu also traverses the parse tree searching for exceptions

    def assign_parents(self, ast):
        if ast:
            assign_parents(ast)

    def post_process_ast(self, ast, issues):
        return ast

    def create_token_stream(self, lexer: Lexer) -> TokenStream:
        return CommonTokenStream(lexer)

    @abstractmethod
    def create_antlr_lexer(self, input_stream: InputStream) -> Lexer:
        """Creates the lexer."""
        pass

    @abstractmethod
    def create_antlr_parser(self, token_stream: TokenStream) -> Parser:
        """Creates the first-stage parser."""
        pass

    @abstractmethod
    def parse_tree_to_ast(self, root, consider_range: bool, issues: List[Issue], source: Source) -> Optional[Node]:
        pass

    def attach_listeners(self, recognizer: Recognizer, issues: List[Issue]):
        recognizer.removeErrorListeners()
        recognizer.addErrorListener(ParserErrorListener(issues))


class ParserErrorListener(ErrorListener):
    def __init__(self, issues: List[Issue]):
        self.issues = issues

    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        start_point = Point(line, column)
        end_point = start_point
        if isinstance(offending_symbol, Token):
            end_point = token_end_point(offending_symbol)
        msg = (msg or "unspecified").capitalize()
        self.issues.append(Issue(IssueType.SYNTACTIC, msg, position=Position(start_point, end_point)))
